# CheapDeathmatch release 7 by David Bowland
# ./addons/eventscripts/cheapdeathmatch/cheapdeathmatch.py

# >>> To configure this addon please see cheapdeathmatch.cfg (created whent he addon is first loaded) <<<

"""
Creates fast, deathmatch-style gameplay with only ES 2.0+. Spawn points are created automaticially so CheapDeathmatch is compatible with most maps without setup.
"""


import cfglib
import es
import gamethread
import langlib
import os.path
import playerlib
import random

import psyco
psyco.full()


info = es.AddonInfo()
info.name     = 'CheapDeathmatch'
info.version  = '7'
info.basename = 'cheapdeathmatch'
info.url      = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author   = 'SuperDave'


###


config = cfglib.AddonCFG(es.getAddonPath(info.basename) + '/' + info.basename + '.cfg')

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename':info.basename})
config.text(info.url)
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('\n')
config.text('To configure language settings for %s please see %s_languages.ini' % (info.name, info.basename))
config.text('\n')

config.text('***** General options *****')
cvar_health      = config.cvar('cheapdeathmatch_health',      100, 'Health players receive each spawn')
cvar_lives       = config.cvar('cheapdeathmatch_lives',       2,   'Number of lives players receive at the start of each round--set to 0 for infinite lives')
cvar_lives_steal = config.cvar('cheapdeathmatch_lives_steal', 1,   '0 = players do not receive an extra life for kills, 1 = players receive an extra life for every opponent killed')
cvar_kill_cash   = config.cvar('cheapdeathmatch_kill_cash',   150, 'Amount of cash players receive for killing an opponent')
config.text('\n')

config.text('***** Respawn options *****')
cvar_respawn_attempts    = config.cvar('cheapdeathmatch_respawn_attempts',    50,  'Maximum number of attempts to respawn players')
config.text('If no spawn point is found players will respawn at the place of death.')
cvar_respawn_distance    = config.cvar('cheapdeathmatch_respawn_distance',    600, 'Distance a spawn point must be from opponents for a player to spawn on it')
cvar_respawn_debug       = config.cvar('cheapdeathmatch_respawn_debug',       0,   'Debug level of error message displayed when no spawn point is found for a player')
config.text('Respawn error messages will only be visible if eventscripts_debug is greater than or equal to cheapdeathmatch_respawn_debug.')
config.text('You can find more information on eventscripts_debug here: http://www.eventscripts.com/pages/Eventscripts_debug')
cvar_respawn_create_step = config.cvar('cheapdeathmatch_respawn_create_step', 10,  'Number of seconds between creating spawn points at living player locations--set to 0 to use predefined map spawn points only')
config.text('NOTE: This value _must_ be 0 if the map changes in layout each round (such as cs_havana).')
config.text('\n')

config.text('***** Spawn protection options *****')
config.text('This addon should be compatible with most deathmatch-compatible spawn protection addons if third-party spawn protection is desired.')
cvar_protection_duration    = config.cvar('cheapdeathmatch_protection_duration',    6,  'Number of seconds after spawning players will be invincible')
cvar_protection_alpha_start = config.cvar('cheapdeathmatch_protection_alpha_start', 75, 'Alpha (visibility) of players at spawn--255 is fully visible')
cvar_protection_alpha_steps = config.cvar('cheapdeathmatch_protection_alpha_steps', 3,  'Number of steps to fade invincible players to full visibility--set to 1 for no fade')
config.text('\n')

config.text('***** Anouncement options *****')
cvar_announce_lkd_trigger            = config.cvar('cheapdeathmatch_announce_lkd_trigger',            '!dmlkd', 'Say command to display personal lives, kills, deaths information--set to "none" to remove')
cvar_announce_lkd_trigger_roundstart = config.cvar('cheapdeathmatch_announce_lkd_trigger_roundstart', 2,        '0 = no change, 1 = announce command to trigger deathmatch lives, kills, deaths information on the console each round, 2 = announce command to trigger deathmatch lives, kills, deaths information in chat area each round')
cvar_announce_lkd_chat               = config.cvar('cheapdeathmatch_announce_lkd_chat',               0,        '0 = no change, 1 = announce lives, kills, and deaths to players on the console each spawn, 2 = announce lives, kills, and deaths to players in chat area each spawn')
cvar_announce_lkd_center             = config.cvar('cheapdeathmatch_announce_lkd_center',             1,        '0 = no change, 1 = announce lives, kills, and deaths to players in center text each spawn')
cvar_announce_lkd_hud                = config.cvar('cheapdeathmatch_announce_lkd_hud',                1,        '0 = no change, 1 = announce lives, kills, and deaths to players in a HUD message each spawn')
config.text('')
cvar_announce_stats_trigger            = config.cvar('cheapdeathmatch_announce_stats_trigger',            '!dmstats', 'Say command to display deathmatch statistics--set to "none" to remove')
cvar_announce_stats_trigger_roundstart = config.cvar('cheapdeathmatch_announce_stats_trigger_roundstart', 2,          '0 = no change, 1 = announce command to trigger deathmatch statistics on the console each round, 2 = announce command to trigger deathmatch statistics in chat area each round')
cvar_announce_stats_trigger_rounddeath = config.cvar('cheapdeathmatch_announce_stats_trigger_rounddeath', 2,          '0 = no change, 1 = announce command to trigger deathmatch statistics on the console to players who lose their last life, 2 = announce command to trigger deathmatch statistics in chat area to players who lose their last life')
cvar_announce_stats                    = config.cvar('cheapdeathmatch_announce_stats',                    2,          '0 = announce deathmatch statistics on the console, 1 = announce deathmatch statistics in a menu, 2 = announce deathmatch statistics on the console and in a menu')
cvar_announce_stats_menu               = config.cvar('cheapdeathmatch_announce_stats_menu',               15,         'Number of seconds to display menu of deathmatch statistics--set to 0 for no time limit')
config.text('\n')

config.text('***** Command documentation *****')
config.text('')
config.text('The following server command can be used to respawn _LIVING_ players:')
config.text('')
config.text('cheapdeathmatch_respawn <userid/name/"Steam ID"> [attacker userid/name/"Steam ID"] [weapon] [headshot]')
config.text('   attacker - the attacker to credit for the death of the respawning player')
config.text('      - If omitted, attacker defaults to 0')
config.text('   weapon - the weapon used to kill the respawning player')
config.text('      - If omitted, weapon defaults to "world"')
config.text('   headshot - 0 = player was not killed with a headshot, 1 = player was killed with a headshot')
config.text('      - If omitted, headshot defaults to 0')
config.text('\n')
config.text('In addition to the server command CheapDeathmatch offers a Python command. In order to utilize the Python command CheapDeathmatch must be loaded')
config.text('before your addon and your addon must contain the import line:')
config.text('')
config.text('from cheapdeathmatch import cheapdeathmatch')
config.text('')
config.text('You may then use the following command:')
config.text('')
config.text('cheapdeathmatch.respawn(int_userid, int_attacker=0, str_weapon=\'world\', int_headshot=0)')
config.text('- Respawns a living player')
config.text('   int_userid - integer containing the userid of the player to respawn')
config.text('   int_attacker - integer containing the userid of the player to credit for the death of the respawning player')
config.text('   str_weapon - string containing the weapon used to kill the respawning player')
config.text('   int_headshot - 0 = player was not killed with a headshot, 1 = player was killed with a headshot')

config.write()


###


class SpawnPointManager(object):
   friend_dist = 100

   def __init__(self, maxcount=1000):
      self.locations = []
      self.surefires = []
      self.obstacles = {}
      self.maxcount  = maxcount

   """ Spawn point creation functions """

   def mapStart(self):
      self.removeDelay()

      self.locations = []
      self.obstacles.clear()

      for index in es.createentitylist('info_player_terrorist').keys() + es.createentitylist('info_player_counterterrorist').keys():
         self._addLocation(map(float, es.getindexprop(index, 'CBaseEntity.m_vecOrigin').split(',')))

      self.surefires = list(self.locations)

   def roundStart(self):
      self.removeDelay()

      delay = float(cvar_respawn_create_step)
      if delay and len(self.locations) <= self.maxcount:
         gamethread.delayedname(delay, 'cheapdeathmatch_respawn_create', self._createSpawnPoints)

   def removeDelay(self):
      gamethread.cancelDelayed('cheapdeathmatch_respawn_create')

   def _addLocation(self, location):
      if location in self.locations or len(self.locations) >= self.maxcount: return

      self.locations.append(location)

   def _createSpawnPoints(self):
      for userid in filter(lambda x: not dm.getPlayer(x).isDead(), es.getUseridList()):
         if es.getplayerprop(userid, 'CBasePlayer.m_fFlags') & 1 and not es.getplayerprop(userid, 'CBasePlayer.localdata.m_Local.m_bDucked') & 1:
            self._addLocation(es.getplayerlocation(userid))

      delay = float(cvar_respawn_create_step)
      if delay and len(self.locations) <= self.maxcount:
         gamethread.delayedname(delay, 'cheapdeathmatch_respawn_create', self._createSpawnPoints)

   """ Spawn point retrieval functions """

   def getLocation(self, userid, surefire=False):
      spawns = list(self.surefires if surefire else self.locations)
      if not spawns: return None

      team       = es.getplayerteam(userid)
      enemy_dist = int(cvar_respawn_distance)
      players    = self.obstacles.copy()

      for loop_userid in filter(lambda x: not dm.getPlayer(x).isDead(), es.getUseridList()):
         players[es.getplayerlocation(loop_userid)] = self.friend_dist if es.getplayerteam(loop_userid) == team and loop_userid <> userid else enemy_dist

      if team == 3:
         for index in es.createentitylist('hostage_entity'):
            players[self.getEntityLocation(index)] = enemy_dist
         for index in es.createentitylist('planted_c4'):
            players[self.getEntityLocation(index)] = enemy_dist

      elif team == 2:
         for index in es.createentitylist('weapon_c4'):
            players[self.getEntityLocation(index)] = enemy_dist

      for count in xrange(cvar_respawn_attempts):
         location = random.choice(spawns)
         if self._checkLocation(location, players):
            x, y, z = location
            return x, y, z + 4

         spawns.remove(location)
         if not spawns: break

      return None

   def _checkLocation(self, location, players):
      for player_location in players:
         if sum((location[x] - player_location[x]) ** 2 for x in range(3)) ** 0.5 <= players[player_location]: return False

      return True

   def addObstacle(self, location):
      self.obstacles[location] = self.friend_dist

   def addSurefire(self, location):
      if location in self.surefires: return

      self.surefires.append(location)

   @staticmethod
   def getEntityLocation(index):
      return tuple(map(float, es.getindexprop(index, 'CBaseEntity.m_vecOrigin').split(',')))

spawnpoints = SpawnPointManager()


###


class DeathmatchPlayer(object):
   lives     = 0
   kills     = 0
   deaths    = 0
   protected = False

   fired_spawn = False
   fired_death = False

   def __init__(self, userid):
      self.userid = userid
      self.handle = es.getplayerhandle(userid)

   def removeDelays(self):
      gamethread.cancelDelayed('cheapdeathmatch_protection_%s' % self.userid)
      gamethread.cancelDelayed('cheapdeathmatch_respawn_%s' % self.userid)
      gamethread.cancelDelayed('cheapdeathmatch_firespawn_%s' % self.userid)

   """ Event functions """

   def roundStart(self):
      trigger_option = int(cvar_announce_lkd_trigger_roundstart)
      if trigger_option and lkd_trigger:
         if trigger_option == 1:
            es.cexec(self.userid, 'echo ' + self.__removeTags(lang_text('lkd trigger', {'trigger':lkd_trigger}, self.getLang())))
         else:
            es.tell(self.userid, '#multi', lang_text('lkd trigger', {'trigger':lkd_trigger}, self.getLang()))

      trigger_option = int(cvar_announce_stats_trigger_roundstart)
      if trigger_option and stats_trigger:
         if trigger_option == 1:
            es.cexec(self.userid, 'echo ' + self.__removeTags(lang_text('stats trigger', {'trigger':stats_trigger}, self.getLang())))
         else:
            es.tell(self.userid, '#multi', lang_text('stats trigger', {'trigger':stats_trigger}, self.getLang()))

   def playerSpawn(self):
      if not self.fired_spawn:
         self.lives = int(cvar_lives)
         self.removeDelays()
      self.fired_spawn = False

      protection = float(cvar_protection_duration)
      if protection:
         self.protected = True
         self.setHealth(1024 + cvar_health)
         self.setHitbox(False)

         alpha = int(cvar_protection_alpha_start)
         if alpha == 255:
            gamethread.delayedname(protection, 'cheapdeathmatch_protection_%s' % self.userid, self._protectionStep, (0., 1, 1, alpha))

         else:
            es.server.queuecmd('es_xfire %s !self addoutput \"rendermode 1\"' % self.userid)
            es.server.queuecmd('es_xfire %s !self alpha %s' % (self.userid, alpha))

            steps = float(cvar_protection_alpha_steps)
            delay = protection / steps
            gamethread.delayedname(delay, 'cheapdeathmatch_protection_%s' % self.userid, self._protectionStep, (delay, 1, steps, alpha))

      else:
         self.protected = False
         self.setHealth((0 if self.lives == 1 else 1024) + cvar_health)
         self.setHitbox(True)

      self.announceLKD()

   def playerDeath(self, attacker):
      self.deaths += 1

      if attacker:
         attacker_player = dm.getPlayer(attacker)
         attacker_player.playerKill()

      if self.fired_death:
         self.fired_death = False
         return

      self.removeDelays()

      if attacker:
         attacker_player.killCash()

      trigger_option = int(cvar_announce_stats_trigger_rounddeath)
      if trigger_option and stats_trigger:
         if trigger_option == 1:
            es.cexec(self.userid, 'echo ' + self.__removeTags(lang_text('stats trigger', {'trigger':stats_trigger}, self.getLang())))
         else:
            es.tell(self.userid, '#multi', lang_text('stats trigger', {'trigger':stats_trigger}, self.getLang()))

   def playerKill(self):
      self.kills += 1

      if not int(cvar_lives_steal): return

      self.lives += 1

      health = self.getHealth()
      if health < 1025:
         self.setHealth(health + 1024)

   def killCash(self, default=300):
      es.setplayerprop(self.userid, 'CCSPlayer.m_iAccount', max(16000, es.getplayerprop(self.userid, 'CCSPlayer.m_iAccount') + cvar_kill_cash - default))

   def playerHurt(self, attacker, weapon, hitgroup):
      if self.protected:
         self.setHealth(1024 + cvar_health)
         es.setplayerprop(self.userid, 'CCSPlayer.cslocaldata.m_flVelocityModifier', 1)
         return

      if attacker:
         dm.getPlayer(attacker).removeProtection()

      if self.lives == 1 or self.isDead(): return

      if self.getHealth() >= 1025: return

      self.setHealth(2048)
      self.respawn(attacker, weapon, hitgroup == 1)

   """ Respawn functions """

   def respawn(self, attacker=0, weapon='world', headshot=False):
      self.fired_death = True
      es.event('initialize', 'player_death')
      es.event('setint',     'player_death', 'userid',   self.userid)
      es.event('setint',     'player_death', 'attacker', attacker)
      es.event('setstring',  'player_death', 'weapon',   weapon)
      es.event('setint',     'player_death', 'headshot', int(headshot))
      es.event('fire',       'player_death')

      for index in es.createentitylist('weapon_c4'):
         if es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity') == self.handle:
            es.server.queuecmd('es_xgive %s weapon_c4' % self.userid)
            #es.sexec(self.userid, 'use weapon_c4')
            #es.sexec(self.userid, 'drop')

      if attacker:
         attacker_player = dm.getPlayer(attacker)
         attacker_player.killCash(0)
         attacker_player.addScore(-1 if es.getplayerteam(attacker) == es.getplayerteam(self.userid) else 1)

      if self.lives: self.lives -= 1
      self.removeDelays()
      self._newLocation()

      gamethread.delayedname(0.2, 'cheapdeathmatch_firespawn_%s' % self.userid, self._fireSpawn)

   def _fireSpawn(self):
      self.fired_spawn = True
      es.event('initialize', 'player_spawn')
      es.event('setint',     'player_spawn', 'userid', self.userid)
      es.event('fire',       'player_spawn')

      for index in es.createentitylist('weapon_c4'):
         if es.getindexprop(index, 'CBaseEntity.m_hOwnerEntity') == self.handle:
            es.server.cmd('es_xremove %s' % index)

   def _newLocation(self, location=None, count=0):
      if not location is None:
         current_location = es.getplayerlocation(self.userid)
         if sum((location[x] - current_location[x]) ** 2 for x in range(3)) ** 0.5 > 2:
            spawnpoints.addSurefire(location)
            return

         es.dbgmsg(1, 'CheapDeathmatch: Found player %s (userid %s) stuck at %s' % (self.getName(), self.userid, current_location))
         spawnpoints.addObstacle(current_location)

         if count > 20:
            es.dbgmsg(cvar_respawn_debug, 'CheapDeathmatch: Unable to free player %s (userid %s) stuck at %s' % (self.getName(), self.userid, current_location))
            return

      location = spawnpoints.getLocation(self.userid, count > 15)
      if location is None:
         es.dbgmsg(cvar_respawn_debug, 'CheapDeathmatch: Unable to find spawn point for %s (userid %s)' % (self.getName(), self.userid))
         return

      es.server.queuecmd('es_xsetpos %s %s %s %s' % ((self.userid,) + location))
      gamethread.cancelDelayed('cheapdeathmatch_respawn_%s' % self.userid)
      gamethread.delayedname(0.3, 'cheapdeathmatch_respawn_%s' % self.userid, self._newLocation, (location, count + 1))

   """ Get/set functions """

   def isDead(self):
      return es.getplayerprop(self.userid, 'CBasePlayer.pl.deadflag')

   def getName(self):
      return es.getplayername(self.userid)

   def getHealth(self):
      return es.getplayerprop(self.userid, 'CBasePlayer.m_iHealth')

   def setHealth(self, health):
      es.setplayerprop(self.userid, 'CBasePlayer.m_iHealth', health)

   def setHitbox(self, value):
      es.setplayerprop(self.userid, 'CBaseAnimating.m_nHitboxSet', 0 if value else 2)

   def getLang(self):
      return playerlib.getPlayer(self.userid).get('lang')

   def addScore(self, amount):
      entity_list = es.createentitylist('game_score')
      if len(entity_list) > 1:
         for index in entity_list.keys()[1:]:
            es.server.cmd('es_xremove %s' % index)
      elif not entity_list:
         es.server.cmd('es_xentcreate %s game_score' % self.userid)

      es.server.queuecmd('es_xfire %s game_score addoutput \"spawnflags 1\"' % self.userid)
      es.server.queuecmd('es_xfire %s game_score addoutput \"points %s\"' % (self.userid, amount))
      es.server.queuecmd('es_xfire %s game_score applyscore' % self.userid)

   """ Spawn protection functions """

   def _protectionStep(self, delay, step, totalsteps, alpha):
      if step >= totalsteps:
         if alpha <> 255:
            es.server.queuecmd('es_xfire %s !self alpha 255' % self.userid)

         self.protected = False
         self.setHealth((0 if self.lives == 1 else 1024) + cvar_health)
         self.setHitbox(True)

      else:
         es.server.queuecmd('es_xfire %s !self alpha %s' % (self.userid, ((255 - alpha) * step) / totalsteps + alpha))
         gamethread.delayedname(delay, 'cheapdeathmatch_protection_%s' % self.userid, self._protectionStep, (delay, step + 1, totalsteps, alpha))

   def removeProtection(self):
      if not self.protected: return
      self.protected = False

      self.setHealth((0 if self.lives == 1 else 1024) + cvar_health)
      self.setHitbox(True)
      es.server.queuecmd('es_xfire %s !self alpha 255' % self.userid)
      gamethread.cancelDelayed('cheapdeathmatch_protection_%s' % self.userid)

   """ Stats functions """

   def announceLKD(self):
      tokens      = {'lives':self.lives, 'kills':self.kills, 'deaths':self.deaths}
      chat_option = int(cvar_announce_lkd_chat)
      text        = lang_text('lkd' if self.lives and not self.isDead() else 'kd', tokens, self.getLang())

      if chat_option:
         if chat_option == 1:
            es.cexec(self.userid, 'echo ' + self.__removeTags(text))
         else:
            es.tell(self.userid, '#multi', text)

      if int(cvar_announce_lkd_center):
         es.centertell(self.userid, self.__removeTags(text))

      if int(cvar_announce_lkd_hud):
         self._hudHint(self.__removeTags(text))

   def showStats(self):
      player_lang = self.getLang()
      announce    = int(cvar_announce_stats)
      menu_text   = lang_text('stats header', {}, player_lang)

      if announce <> 1:
         es.cexec(self.userid, 'echo \"%s\"' % menu_text)
      menu_text += '\n \n'

      for count, player in enumerate(sorted(map(dm.getPlayer, filter(lambda x: es.getplayerteam(x) in (2, 3), es.getUseridList())), key=lambda x: (~x.kills, ~x.lives, x.deaths, x.getName().lower()))):
         line = str(count + 1) + '. ' + lang_text('stats line lkd' if player.lives and not player.isDead() else 'stats line kd', {'player':player.getName(), 'kills':player.kills, 'lives':player.lives, 'deaths':player.deaths}, player_lang)

         if announce <> 1:
            es.cexec(self.userid, 'echo \"%s\"' % line)
         menu_text += ('->' if player is self else '') + line + '\n'

      if announce:
         popup.send(self.userid, float(cvar_announce_stats_menu), menu_text[:menu_text.rfind('\n', 0, 487)])

   """ Internal functions """

   def _hudHint(self, message):
      """Sends the player a HUD hint"""
      es.usermsg('create', 'cheapdeathmatch_hud', 'HintText')
      es.usermsg('write', 'short', 'cheapdeathmatch_hud', -1)
      es.usermsg('write', 'string', 'cheapdeathmatch_hud', message)
      es.usermsg('send', 'cheapdeathmatch_hud', self.userid)
      es.usermsg('delete', 'cheapdeathmatch_hud')

   @staticmethod
   def __removeTags(text):
      return text.replace('#lightgreen', '').replace('#green', '').replace('#default', '')

###


class DeathmatchManager(object):
   def __init__(self):
      self.players = {}

   def getPlayer(self, userid):
      userid = int(userid)
      if userid not in self.players:
         self.players[userid] = DeathmatchPlayer(userid)

      return self.players[userid]

   def removePlayer(self, userid):
      userid = int(userid)
      if userid in self.players:
         self.players[userid].removeDelays()
         del self.players[userid]

   def clearPlayers(self, unload=False):
      for userid in self.players:
         if unload:
            self.players[userid].removeProtection()
         self.players[userid].removeDelays()

      self.players.clear()

dm = DeathmatchManager()


###


def getLangText():
   lang_path = es.getAddonPath(info.basename) + '/%s_languages.ini' % info.basename
   if os.path.isfile(lang_path):
      return langlib.Strings(lang_path)

   else:
      es.dbgmsg(0, '%(name)s: Unable to load %(basename)s_languages.ini! Please ensure it is in the ./%(basename)s/ directory.' % {'name':info.name, 'basename':info.basename})
      return lambda x, y, z: '%s: %s_languages.ini not found' % (info.name, info.basename)

lang_text = getLangText()


###


stats_trigger = None
lkd_trigger   = None


def load():
   """
   Registers the server command
   Ensures critical server variables are created by the config
   Registers the say command
   Calls es_map_start
   """
   global stats_trigger
   global lkd_trigger

   config.execute()

   temptrigger = str(cvar_announce_stats_trigger)
   if temptrigger.lower() <> 'none' and not es.exists('saycommand', temptrigger):
      es.regsaycmd(temptrigger, 'cheapdeathmatch/stats_cmd', 'Say command to display deathmatch stats')
      stats_trigger = temptrigger

   temptrigger = str(cvar_announce_lkd_trigger)
   if temptrigger.lower() <> 'none' and not es.exists('saycommand', temptrigger):
      es.regsaycmd(temptrigger, 'cheapdeathmatch/lkd_cmd', 'Say command to display deathmatch lives, kills, deaths infomation')
      lkd_trigger = temptrigger

   for userid in filter(lambda x: not dm.getPlayer(x).isDead(), es.getUseridList()):
     dm.getPlayer(userid).playerSpawn()
   spawnpoints.mapStart()


def es_map_start(event_var):
   spawnpoints.mapStart()

   dm.clearPlayers()
   popup.clearPlayers()


def round_start(event_var):
   """
   Clears active menu delays
   Sets starting lives for players
   Announces stats trigger
   """
   spawnpoints.roundStart()

   for userid in es.getUseridList():
      dm.getPlayer(userid).roundStart()

   popup.clearPlayers()


def player_spawn(event_var):
   if event_var['es_userteam'] not in ('2', '3'): return

   dm.getPlayer(event_var['userid']).playerSpawn()


def player_hurt(event_var):
   dm.getPlayer(event_var['userid']).playerHurt(int(event_var['attacker']), event_var['weapon'], int(event_var['hitgroup']))


def player_death(event_var):
   dm.getPlayer(event_var['userid']).playerDeath(int(event_var['attacker']))


def player_disconnect(event_var):
   userid = event_var['userid']

   dm.removePlayer(userid)
   popup.removePlayer(userid)


def unload():
   """
   Unregisters say trigger
   """
   spawnpoints.removeDelay()

   dm.clearPlayers(True)
   popup.clearPlayers(True)

   if not stats_trigger is None:
      es.unregsaycmd(stats_trigger)
   if not lkd_trigger is None:
      es.unregsaycmd(lkd_trigger)

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)


###


def respawn_cmd():
   """
   cheapdeathmatch_respawn <userid/name/\"Steam ID\"> [attacker userid/name/\"Steam ID\"] [weapon] [headshot]
   """
   argc = es.getargc()
   if argc in (2, 3, 4, 5):
      player = es.getargv(1)
      userid = es.getuserid(player)
      if userid:
         if argc >= 3:
            player   = es.getargv(2)
            attacker = es.getuserid(player)
            if not attacker:
               es.dbgmsg(0, 'CheapDeathmatch: Unable to find attacker \"%s\"' % player)
               return

         else:
            attacker = 0

         dm.getPlayer(userid).respawn(attacker, es.getargv(3).lower() if argc >= 4 else 'world', (es.getargv(4) <> '0') if argc >= 5 else False)

      else:
         es.dbgmsg(0, 'CheapDeathmatch: Unable to find player \"%s\"' % player)

   else:
      es.dbgmsg(0, 'Syntax: cheapdeathmatch_respawn <userid/name/\"Steam ID\"> [attacker userid/name/\"Steam ID\"] [weapon] [headshot]')

if not es.exists('command', 'cheapdeathmatch_respawn'):
   es.regcmd('cheapdeathmatch_respawn', 'cheapdeathmatch/respawn_cmd', 'cheapdeathmatch_respawn <userid/name/\"Steam ID\"> [attacker userid/name/\"Steam ID\"] [weapon] [headshot]\nServer command to respawn living players')


###


def lkd_cmd():
   dm.getPlayer(es.getcmduserid()).announceLKD()


def stats_cmd():
   dm.getPlayer(es.getcmduserid()).showStats()


###


class CheapPopup(object):
   class Player(object):
      menu     = ()
      callback = None
      timeout  = None

      def __init__(self, userid, parent):
         self.userid = userid
         self.parent = parent

      def send(self, duration, text, callback=None, timeout=None):
         if not self.menu:
            es.menu(0, self.userid, ' ')
            es.cexec(self.userid, 'slot10')

         self.menu     = (duration, self.userid, text)
         self.callback = callback
         self.timeout  = timeout

      def unSend(self):
         if not self.menu:
            es.cexec(self.userid, 'slot10')

         self.removeDelay()
         self.parent.removePlayer(self.userid)

      def selection(self, item):
         if item == 10 and self.menu:
            es.menu(*self.menu)

            delay = self.menu[0]
            if delay:
               gamethread.delayedname(delay, self.parent.basename + '_popup_' + str(self.userid), self._endDelay)

            self.menu = ()
            return

         self.parent.removePlayer(self.userid)

         if self.callback(self.userid, item) if self.callback else True:
            if item <> 10:
               es.cexec(self.userid, 'slot%s' % item)

      def _endDelay(self):
         self.parent.removePlayer(self.userid)
         if self.timeout:
            self.timeout(self.userid)

      def removeDelay(self):
         gamethread.cancelDelayed(self.parent.basename + '_popup_' + str(self.userid))

   """ Begin CheapPopup """
   basename = __name__

   def __init__(self):
      self.players = {}

   """ Player functions """

   def getPlayer(self, userid):
      userid = int(userid)
      if userid not in self.players:
         self.players[userid] = self.Player(userid, self)
         es.addons.registerClientCommandFilter(self._selection_hook)

      return self.players[userid]

   def removePlayer(self, userid):
      userid = int(userid)
      if userid in self.players:
         self.players[userid].removeDelay()
         del self.players[userid]

   def clearPlayers(self, unsend=False):
      if unsend:
         for userid in self.players.keys():
            self.players[userid].unSend()

      else:
         for userid in self.players:
            self.players[userid].removeDelay()

         self.players.clear()

      es.addons.unregisterClientCommandFilter(self._selection_hook)

   """ Popup functions """

   def send(self, userid, duration, text, callback=None, timeout=None):
      self.getPlayer(userid).send(duration, text, callback, timeout)

   def unsend(self, userid):
      userid = int(userid)
      if userid in self.players:
         self.getPlayer(userid).unSend()

   def _selection_hook(self, userid, args):
      if userid in self.players and args[0].lower() == 'menuselect' and len(args) > 1:
         self.getPlayer(userid).selection(int(args[1]))

      return True

popup = CheapPopup()